/*
 * The Apache Software License, Version 1.1li
 *
 * Copyright (c) 2002 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks;

import org.apache.tools.ant.types.FlexInteger;
import net.sf.antcontrib.cpptasks.types.FlexLong;
import net.sf.antcontrib.cpptasks.compiler.LinkerConfiguration;
import net.sf.antcontrib.cpptasks.types.LibrarySet;
import net.sf.antcontrib.cpptasks.types.SystemLibrarySet;
import net.sf.antcontrib.cpptasks.types.LinkerArgument;
import net.sf.antcontrib.cpptasks.gcc.GccLinker;
import net.sf.antcontrib.cpptasks.borland.BorlandLinker;
import net.sf.antcontrib.cpptasks.compiler.*;
import org.apache.tools.ant.Project;
import java.io.File;
import org.apache.tools.ant.BuildException;
import net.sf.antcontrib.cpptasks.devstudio.DevStudioLinker;
import net.sf.antcontrib.cpptasks.compaq.CompaqVisualFortranLinker;
import java.util.*;

/**
 * A linker definition.  linker elements may be placed
 *    either as children of a cc element or the project element.
 *    A linker element with an id attribute may be referenced
 *    by linker elements with refid or extends attributes.
 *
 * @author Adam Murdoch
 * @author Curt Arnold
 */
public class LinkerDef extends ProcessorDef
{
    private final Vector librarySets = new Vector();
    private final Vector sysLibrarySets = new Vector();
    private long base;
    private int stack;
    private Boolean map;
    private Boolean fixed;
    private String entry;
    private Boolean incremental;

    public LinkerDef() {
        base = -1;
        stack = -1;
    }


    /**
     * Sets linker type.
     *
     *
     * <table width="100%" border="1">
     *    <thead>Supported linkers</thead>
        *       <tr><td>gcc</td><td>Gcc Linker</td></tr>
        *       <tr><td>ld</td><td>Ld Linker</td></tr>
        *       <tr><td>ar</td><td>Gcc Librarian</td></tr>
        *       <tr><td>msvc</td><td>Microsoft Linker</td></tr>
        *       <tr><td>bcc</td><td>Borland Linker</td></tr>
        *       <tr><td>df</td><td>Compaq Visual Fortran Linker</td></tr>
        *       <tr><td>icl</td><td>Intel Linker for Windows (IA-32)</td></tr>
        *       <tr><td>ecl</td><td>Intel Linker for Windows (IA-64)</td></tr>
        *       <tr><td>icc</td><td>Intel Linker for Linux (IA-32)</td></tr>
        *       <tr><td>ecc</td><td>Intel Linker for Linux (IA-64)</td></tr>
        *       <tr><td>CC</td><td>Sun ONE Linker</td></tr>
        *       <tr><td>aCC</td><td>HP aC++ Linker</td></tr>
        *       <tr><td>os390</td><td>OS390 Linker</td></tr>
        *       <tr><td>os390batch</td><td>OS390 Linker</td></tr>
        *       <tr><td>os400</td><td>IccLinker</td></tr>
        *       <tr><td>sunc89</td><td>C89 Linker</td></tr>
        *       <tr><td>xlC</td><td>VisualAge Linker</td></tr>
     * </table>
     *
     */
    public void setName(LinkerEnum name) throws BuildException {
      if ( isReference() ) {
          throw tooManyAttributes();
      }
      Linker linker = name.getLinker();
      super.setProcessor(linker);
    }

    protected void setProcessor(Processor proc) throws BuildException {
        Linker linker = null;
        if (proc instanceof Linker) {
          linker = (Linker) proc;
        } else {
          LinkType linkType = new LinkType();
          linker = proc.getLinker(linkType);
        }
        super.setProcessor(linker);
    }

    /**
     * Adds a linker command-line arg.
     */
    public void addConfiguredLinkerArg(LinkerArgument arg)
    {
        addConfiguredProcessorArg(arg);
    }

    /**
     * Adds a system library set.
     */
    public void addSyslibset(SystemLibrarySet libset)
    {
        if(isReference()) {
          throw super.noChildrenAllowed();
        }
        if(libset == null) {
          throw new NullPointerException("libset");
        }
        sysLibrarySets.addElement(libset);
    }


    /**
     * Adds a system library set.
     */
    public void addLibset(LibrarySet libset)
    {
        if(isReference()) {
          throw super.noChildrenAllowed();
        }
        if(libset == null) {
          throw new NullPointerException("libset");
        }
        librarySets.addElement(libset);
    }

    public Processor getProcessor() {
      Linker linker = (Linker) super.getProcessor();
      if(linker == null) {
        linker = GccLinker.getInstance();
      }
      if (getLibtool() && linker instanceof CommandLineLinker) {
        CommandLineLinker cmdLineLinker = (CommandLineLinker) linker;
        linker = cmdLineLinker.getLibtoolLinker();
      }
      return linker;
    }

    /**
     *  Sets the base address.  May be specified in either decimal
     *    or hex.
     *
     *  @param base base address
     *
     */
    public void setBase(FlexLong base) {
      if ( isReference() ) {
          throw tooManyAttributes();
      }
      this.base = base.longValue();
    }

    public long getBase(LinkerDef[] defaultProviders, int index) {
      if(isReference()) {
          return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
              getBase(defaultProviders,index);
      }
      if(base <= 0) {
        if(defaultProviders != null && index < defaultProviders.length) {
          return defaultProviders[index].getBase(defaultProviders,index+1);
        }
      }
      return base;
    }

    /**
     * Sets stack size in bytes.
     */
    public void setStack(FlexInteger stack) {
      if ( isReference() ) {
          throw tooManyAttributes();
      }
      this.stack = stack.intValue();
    }

    public int getStack(LinkerDef[] defaultProviders, int index) {
      if(isReference()) {
            return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
                getStack(defaultProviders, index);
      }
      if(stack < 0) {
        if(defaultProviders != null && index < defaultProviders.length) {
          return defaultProviders[index].getStack(defaultProviders,index+1);
        }
      }
      return stack;
    }


    /**
     *  If set to true, a map file will be produced.
     */
    public void setMap(boolean map) {
      if ( isReference() ) {
          throw tooManyAttributes();
      }
      this.map = booleanValueOf(map);
    }

    /**
     *  If true, allows incremental linking.
     *
     */
    public void setIncremental(boolean incremental) {
      if ( isReference() ) {
          throw tooManyAttributes();
      }
      this.map = booleanValueOf(incremental);
    }

    /**
     *  If true, marks the file to be loaded only at its preferred address.
     */
    public void setFixed(boolean fixed) {
      if ( isReference() ) {
          throw tooManyAttributes();
      }
      this.fixed = booleanValueOf(fixed);
    }


    /**
     * Sets the starting address.
     *
     * @param name function name
     */
    public void setEntry(String entry) {
      if ( isReference() ) {
          throw tooManyAttributes();
      }
      this.entry = entry;
    }

    public Boolean getFixed(LinkerDef[] defaultProviders, int index) {
      if(isReference()) {
            return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
                getFixed(defaultProviders,index);
      }
      if(fixed == null) {
        if(defaultProviders != null && index < defaultProviders.length) {
          return defaultProviders[index].getFixed(defaultProviders,index+1);
        }
      }
      return fixed;
    }

    public boolean getMap(LinkerDef[] defaultProviders, int index) {
      if(isReference()) {
            return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
                getMap(defaultProviders,index);
      }
      if(map != null) {
        return map.booleanValue();
      }
      if(defaultProviders != null && index < defaultProviders.length) {
        defaultProviders[index].getMap(defaultProviders,index+1);
      }
      return false;
    }

    public boolean getIncremental(LinkerDef[] defaultProviders, int index) {
      if(isReference()) {
            return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
                getIncremental(defaultProviders,index);
      }
      if(incremental != null) {
        return incremental.booleanValue();
      }
      if(defaultProviders != null && index < defaultProviders.length) {
        defaultProviders[index].getIncremental(defaultProviders,index+1);
      }
      return false;
    }


    /**
     * Returns an array of active library sets for this linker definition.
     */
    public LibrarySet[] getActiveLibrarySets(LinkerDef[] defaultProviders, int index) {
      if (isReference()) {
            return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
                getActiveUserLibrarySets(defaultProviders,index);
      }
      Vector libsets = new Vector();
      for(int i = index; i < defaultProviders.length; i++) {
        defaultProviders[i].addActiveUserLibrarySets(project,libsets);
        defaultProviders[i].addActiveSystemLibrarySets(project,libsets);
      }
      addActiveUserLibrarySets(project,libsets);
      addActiveSystemLibrarySets(project,libsets);
      LibrarySet[] sets = new LibrarySet[libsets.size()];
      libsets.copyInto(sets);
      return sets;
    }

    /**
     * Returns an array of active library sets for this linker definition.
     */
    public LibrarySet[] getActiveUserLibrarySets(LinkerDef[] defaultProviders, int index) {
      if (isReference()) {
            return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
                getActiveUserLibrarySets(defaultProviders,index);
      }
      Vector libsets = new Vector();
      for(int i = index; i < defaultProviders.length; i++) {
        defaultProviders[i].addActiveUserLibrarySets(project,libsets);
      }
      addActiveUserLibrarySets(project,libsets);
      LibrarySet[] sets = new LibrarySet[libsets.size()];
      libsets.copyInto(sets);
      return sets;
    }

    /**
     * Returns an array of active library sets for this linker definition.
     */
    public LibrarySet[] getActiveSystemLibrarySets(LinkerDef[] defaultProviders, int index) {
      if (isReference()) {
            return ((LinkerDef)getCheckedRef(LinkerDef.class, "LinkerDef")).
                getActiveUserLibrarySets(defaultProviders,index);
      }
      Vector libsets = new Vector();
      for(int i = index; i < defaultProviders.length; i++) {
        defaultProviders[i].addActiveSystemLibrarySets(project,libsets);
      }
      addActiveSystemLibrarySets(project,libsets);
      LibrarySet[] sets = new LibrarySet[libsets.size()];
      libsets.copyInto(sets);
      return sets;
    }

    private void addActiveSystemLibrarySets(Project project, Vector libsets) {
      addActiveLibrarySet(project,libsets,sysLibrarySets);
    }

    private void addActiveUserLibrarySets(Project project, Vector libsets) {
      addActiveLibrarySet(project,libsets,librarySets);
    }

    private void addActiveLibrarySet(Project project, Vector libsets, Vector srcSets) {
      Enumeration srcenum = srcSets.elements();
      while (srcenum.hasMoreElements()) {
        LibrarySet set = (LibrarySet) srcenum.nextElement();
        if (set.isActive(project)) {
          libsets.addElement(set);
        }
      }
    }

    public void visitUserLibraries(Linker linker,FileVisitor libraryVisitor) {
      if(project == null) {
        throw new java.lang.IllegalStateException("project must be set");
      }
      if ( isReference() ) {
            LinkerDef master = ((LinkerDef)getCheckedRef(LinkerDef.class, "Linker"));
            master.visitUserLibraries(linker,libraryVisitor);
      }
      else {
        //
        //   if this linker extends another,
        //      visit its libraries first
        //
        LinkerDef extendsDef = (LinkerDef) getExtends();
        if(extendsDef != null) {
          extendsDef.visitUserLibraries(linker,libraryVisitor);
        }
        //
        //   visit the user libraries
        //
        if (librarySets.size() > 0) {
            File[] libpath = linker.getLibraryPath();
            for(int i = 0; i < librarySets.size(); i++) {
              LibrarySet set = (LibrarySet) librarySets.elementAt(i);
              if(set.isActive(project)) {
                set.visitLibraries(project,linker,libpath, libraryVisitor);
              }
            }
        }
      }
    }


    public void visitSystemLibraries(Linker linker,FileVisitor libraryVisitor) {
      if(project == null) {
        throw new java.lang.IllegalStateException("project must be set");
      }
      if ( isReference() ) {
            LinkerDef master = ((LinkerDef)getCheckedRef(LinkerDef.class, "Linker"));
            master.visitSystemLibraries(linker,libraryVisitor);
      }
      else {
        //
        //   if this linker extends another,
        //      visit its libraries first
        //
        LinkerDef extendsDef = (LinkerDef) getExtends();
        if(extendsDef != null) {
          extendsDef.visitSystemLibraries(linker,libraryVisitor);
        }
        if (sysLibrarySets.size() > 0) {
            File[] libpath = linker.getLibraryPath();
            for (int i = 0; i < sysLibrarySets.size(); i++) {
            LibrarySet set = (LibrarySet) sysLibrarySets.elementAt(i);
            if (set.isActive(project)) {
                set.visitLibraries(project,linker,libpath, libraryVisitor);
            }
          }
        }
      }
    }


     public ProcessorConfiguration createConfiguration(
        CCTask task, LinkType linkType, ProcessorDef baseDef) {
       //
       //    must combine some local context (the linkType)
       //       with the referenced element
       //
       //    get a pointer to the definition (either local or referenced)
       ProcessorDef thisDef = this;
       if(isReference()) {
          thisDef = ((ProcessorDef)getCheckedRef(ProcessorDef.class, "ProcessorDef"));
       }
       //
       //    find the appropriate processor (combines local linkType
       //       with possibly remote linker name)
       Processor proc = getProcessor();
       proc = proc.getLinker(linkType);
 
       ProcessorDef[] defaultProviders = getDefaultProviders(baseDef);
       return proc.createConfiguration(task, linkType, defaultProviders, thisDef);
     }

    public void execute() throws org.apache.tools.ant.BuildException
    {
        throw new org.apache.tools.ant.BuildException("Not an actual task, but looks like one for documentation purposes");
    }

}
