/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2002 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks.borland;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

import net.sf.antcontrib.cpptasks.CUtil;
import net.sf.antcontrib.cpptasks.compiler.CommandLineLinker;
import net.sf.antcontrib.cpptasks.compiler.CommandLineLinkerConfiguration;
import net.sf.antcontrib.cpptasks.compiler.LinkType;
import net.sf.antcontrib.cpptasks.compiler.Linker;


/**
 * Adapter for the Borland(r) ilink32 linker
 *
 * @author Curt Arnold
 */
public final class BorlandLinker extends CommandLineLinker
{
    private static final BorlandLinker instance =
        new BorlandLinker(".exe");
    private static final BorlandLinker dllLinker =
        new BorlandLinker(".dll");

    private BorlandLinker(String outputSuffix)
    {
        super("ilink32", "/bogus", new String[] { ".obj", ".lib",".res" },
          new String[] { ".map", ".pdb", ".lnk" }, outputSuffix,
          false, null);
    }

    public static BorlandLinker getInstance() {
      return instance;
    }

    public String[] getOutputFileSwitch(String outFile) {
      return BorlandProcessor.getOutputFileSwitch(outFile);
    }

    public File[] getLibraryPath() {
      return BorlandProcessor.getEnvironmentPath("ilink32",'L',
        new String[] { "..\\lib" } );
    }




    protected void addFixed(Boolean fixed, Vector args) {
    }


    protected void addMap(boolean map, Vector args) {
      if(!map) {
          args.addElement("-x");
      }
    }

    protected void addBase(long base, Vector args) {
      if(base >= 0) {
        String baseAddr = Long.toHexString(base);
        args.addElement("-b:" + baseAddr);
      }
    }

    protected void addStack(int stack, Vector args) {
      if(stack >= 0) {
        String stackStr = Integer.toHexString(stack);
        args.addElement("-S:" + stackStr);
      }
    }


    public Linker getLinker(LinkType type) {
      if (type.isStaticLibrary()) {
        return BorlandLibrarian.getInstance();
      }
      if (type.isSharedLibrary()) {
        return dllLinker;
      }
      return instance;
    }

    protected String getStartupObject(LinkType linkType) {
      if(linkType.isSharedLibrary()) {
        return "c0d32.obj";
      }
      if(linkType.isSubsystemGUI()) {
        return "c0w32.obj";
      }
      if(linkType.isSubsystemConsole()) {
        return "c0x32.obj";
      }
      return null;
    }

    protected void addImpliedArgs(boolean debug, LinkType linkType, Vector args) {
      if(linkType.isExecutable()) {
        if(linkType.isSubsystemConsole()) {
          args.addElement("/ap");
        } else {
          if(linkType.isSubsystemGUI()) {
            args.addElement("/Tpe");
          }
        }
      }
      if(linkType.isSharedLibrary()) {
        args.addElement("/Tpd");
      }
    }

    public String getCommandFileSwitch(String commandFile) {
      return "@" + commandFile;
    }


    public int getMaximumCommandLength() {
      return 1024;
    }


    /**
     * Prepares argument list for exec command.
     *
     * @param outputFile linker output file
     * @param sourceFiles linker input files (.obj, .o, .res)
     * @param args linker arguments
     * @return arguments for runTask
     */
    protected String[] prepareArguments(String outputDir,
        String outputName,
        String[] sourceFiles, CommandLineLinkerConfiguration config) {
        String[] preargs = config.getPreArguments();
        String[] endargs = config.getEndArguments();
        Vector execArgs = new Vector(preargs.length + endargs.length + 10 + sourceFiles.length);

        execArgs.addElement(this.getCommand());
        for(int i = 0; i < preargs.length; i++) {
          execArgs.addElement(preargs[i]);
        }
        for(int i = 0; i < endargs.length; i++) {
          execArgs.addElement(endargs[i]);
        }

        //
        //  see if the input files have any known startup obj files
        //
        String startup = null;
        for(int i = 0; i < sourceFiles.length; i++) {
          String filename =new File(sourceFiles[i]).getName().toLowerCase();
          if(startup != null &&
            filename.substring(0,2).equals("c0") &&
            filename.substring(3,5).equals("32") &&
            filename.substring(filename.length()-4).equals(".obj")) {
              startup = sourceFiles[i];
          }
        }

        //
        //  c0w32.obj, c0x32.obj or c0d32.obj depending on
        //        link type
        if(startup == null) {
          startup = config.getStartupObject();
        }
        execArgs.addElement(startup);

        Vector resFiles = new Vector();
        Vector libFiles = new Vector();
        String defFile = null;
        StringBuffer buf = new StringBuffer();
        for(int i = 0; i < sourceFiles.length; i++) {
          String last4 = sourceFiles[i].
            substring(sourceFiles[i].length()-4).toLowerCase();
          if(last4.equals(".def")) {
            defFile = quoteFilename(buf,sourceFiles[i]);
          }
          else {
            if(last4.equals(".res")) {
              resFiles.addElement(quoteFilename(buf,sourceFiles[i]));
            }
            else {
              if(last4.equals(".lib")) {
                libFiles.addElement(quoteFilename(buf,sourceFiles[i]));
              }
              else {
                execArgs.addElement(quoteFilename(buf,sourceFiles[i]));
              }
            }
          }
        }

        //
        //   output file name
        //
        String outputFileName = new File(outputDir, outputName).toString();
        execArgs.addElement("," + quoteFilename(buf,outputFileName));

        if(config.getMap()) {
            int lastPeriod = outputFileName.lastIndexOf('.');
            String mapName;
            if(lastPeriod < outputFileName.length() - 4) {
              mapName = outputFileName +".map";
            }
            else {
              mapName = outputFileName.substring(0,lastPeriod) + ".map";
            }
            execArgs.addElement("," + quoteFilename(buf,mapName) + ",");
        }
        else {
            execArgs.addElement(",,");
        }


        //
        //   add all the libraries
        //
        Enumeration libEnum = libFiles.elements();
        while(libEnum.hasMoreElements()) {
          String libName = (String) libEnum.nextElement();
          execArgs.addElement(quoteFilename(buf,libName));
        }

        if(defFile == null) {
          execArgs.addElement(",,");
        }
        else {
          execArgs.addElement("," + quoteFilename(buf,defFile) + ",");
        }

        Enumeration resEnum = resFiles.elements();
        while(resEnum.hasMoreElements()) {
          String resName = (String) resEnum.nextElement();
          execArgs.addElement(quoteFilename(buf,resName));
        }

        String[] execArguments = new String[execArgs.size()];
        execArgs.copyInto(execArguments);
        return execArguments;
    }


     /**
     * Prepares argument list to execute the linker using a
     * response file.
     *
     * @param outputFile linker output file
     * @param args output of prepareArguments
     * @return arguments for runTask
     */
    protected String[] prepareResponseFile(File outputFile,String[] args) throws IOException
    {
        String baseName = CUtil.getBasename(outputFile);
        File commandFile = new File(outputFile.getParent(),baseName + ".lnk");
        FileWriter writer = new FileWriter(commandFile);
        for(int i = 1; i < args.length-1; i++) {
          writer.write(args[i]);
          //
          //  if either the current argument ends with
          //     or next argument starts with a comma then
          //      don't split the line
          if(args[i].endsWith(",") || args[i+1].startsWith(",")) {
            writer.write(' ');
          }
          else {
            //
            //  split the line to make it more readable
            //
            writer.write(" + \n");
          }
        }
        //
        //  write the last argument
        //
        if(args.length > 1) {
          writer.write(args[args.length-1]);
        }
        writer.close();
        String[] execArgs = new String[2];
        execArgs[0] = args[0];
        execArgs[1] = getCommandFileSwitch(commandFile.toString());
        return execArgs;
    }

    public String[] getLibraryPatterns(String[] libnames) {
      return BorlandProcessor.getLibraryPatterns(libnames);
    }


    protected void addIncremental(boolean incremental, Vector args) {
    }

    public boolean isCaseSensitive() {
      return BorlandProcessor.isCaseSensitive();
    }

}
