/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2002 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks.compiler;

import java.io.File;

import net.sf.antcontrib.cpptasks.CCTask;
import net.sf.antcontrib.cpptasks.DependencyInfo;
import org.apache.tools.ant.BuildException;

/**
 * A configuration for a C++ compiler
 *
 * @author Curt Arnold
 */
public final class CommandLineCompilerConfiguration implements CompilerConfiguration
{
    private /* final */ CommandLineCompiler compiler;
    private /* final */ String[] args;
    private /* final */ String identifier;
    private /* final */ boolean rebuild;
    private /* final */ File[] includePath;
    private /* final */ File[] sysIncludePath;
    private String[] endArgs;
    //
    //    include path from environment variable not
    //       explicitly stated in Ant script
    private /* final */ File[] envIncludePath;
    private /* final */ String includePathIdentifier;
    private boolean isPrecompiledHeaderGeneration;
    private String[] exceptFiles;

    public CommandLineCompilerConfiguration(CommandLineCompiler compiler,
      String identifier,
      File[] includePath,
      File[] sysIncludePath,
      File[] envIncludePath,
      String includePathIdentifier,
      String[] args, boolean rebuild,
      String[] endArgs) {
      if(compiler == null) {
        throw new NullPointerException("compiler");
      }
      if(identifier == null) {
        throw new NullPointerException("identifier");
      }
      if(includePathIdentifier == null){
        throw new NullPointerException("includePathIdentifier");
      }
      if(args == null) {
        this.args = new String[0];
      }
      else {
        this.args = (String[]) args.clone();
      }
      if(includePath == null) {
        this.includePath = new File[0];
      }
      else {
        this.includePath = (File[]) includePath.clone();
      }
      if(sysIncludePath == null) {
        this.sysIncludePath = new File[0];
      }
      else {
        this.sysIncludePath = (File[]) sysIncludePath.clone();
      }
      if(envIncludePath == null) {
        this.envIncludePath = new File[0];
      }
      else {
        this.envIncludePath = (File[]) envIncludePath.clone();
      }

      this.compiler = compiler;
      this.rebuild = rebuild;
      this.identifier = identifier;
      this.includePathIdentifier = includePathIdentifier;
      this.endArgs = (String[]) endArgs.clone();
      exceptFiles = null;
      isPrecompiledHeaderGeneration = false;
    }

    public CommandLineCompilerConfiguration(CommandLineCompilerConfiguration base,
        String[] additionalArgs,
        String[] exceptFiles, boolean isPrecompileHeaderGeneration) {
        compiler = base.compiler;
        identifier = base.identifier;
        rebuild = base.rebuild;
        includePath = (File[]) base.includePath.clone();
        sysIncludePath = (File[]) base.sysIncludePath.clone();
        endArgs = (String[]) base.endArgs.clone();
        envIncludePath = (File[]) base.envIncludePath.clone();
        includePathIdentifier = base.includePathIdentifier;
        if (exceptFiles != null) {
          this.exceptFiles = (String[]) exceptFiles.clone();
        }
        this.isPrecompiledHeaderGeneration = isPrecompileHeaderGeneration;

        args = new String[base.args.length + additionalArgs.length];
        for (int i = 0; i < base.args.length; i++) {
          args[i] = base.args[i];
        }
        int index = base.args.length;
        for (int i = 0; i < additionalArgs.length; i++) {
          args[index++] = additionalArgs[i];
        }
    }


    public String toString() {
      return identifier;
    }

    /**
     * Returns a string representation of this configuration.
     * Should be canonical so that equivalent configurations
     * will have equivalent string representations
     */
    public String getIdentifier() {
      return identifier;
    }

    public String getOutputFileName(String inputFile) {
      return compiler.getOutputFileName(inputFile);
    }

    public int bid(String inputFile) {
      int compilerBid = compiler.bid(inputFile);
      if (compilerBid > 0 && exceptFiles != null) {
        for (int i = 0; i < exceptFiles.length; i++) {
          if(inputFile.equals(exceptFiles[i])) {
            return 0;
          }
        }
      }
      return compilerBid;
    }


    public void compile(CCTask task,File outputDir, String[] sourceFiles,
      boolean relentless, ProgressMonitor monitor)
      throws BuildException{
        if (monitor != null) {
          monitor.start(this);
        }
        try {
          compiler.compile(task,outputDir, sourceFiles,args,
            endArgs, relentless, monitor);
          if (monitor != null) {
            monitor.finish(this, true);
          }
        } catch(BuildException ex) {
          if (monitor != null) {
            monitor.finish(this, false);
          }
          throw ex;
        }
    }

    public String getIncludePathIdentifier() {
      return includePathIdentifier;
    }

    public boolean getRebuild() {
      return rebuild;
    }

    public DependencyInfo parseIncludes(CCTask task,File baseDir, File source) {
      return compiler.parseIncludes(task,source,includePath,
          sysIncludePath,envIncludePath,baseDir,
          getIncludePathIdentifier());
    }

    /**
     *
     * This method may be used to get two distinct compiler configurations,
     * one for compiling the specified file and producing a precompiled
     * header file, and a second for compiling other files using the
     * precompiled header file.
     *
     * The last (preferrably only) include directive in the prototype
     * file will be used to mark the boundary between pre-compiled and
     * normally compiled headers.
     *
     * @param prototype A source file (for example, stdafx.cpp) that
     * is used to build the precompiled header file.
     * @returns null if precompiled headers are not supported or a two
     * element array containing the precompiled header generation configuration
     * and the consuming configuration
     *
     */
    public CompilerConfiguration[] createPrecompileConfigurations(
        File prototype, String[] nonPrecompiledFiles) {
       if( compiler instanceof PrecompilingCompiler ) {
          return ((PrecompilingCompiler) compiler).
            createPrecompileConfigurations(this, prototype,
                nonPrecompiledFiles);
       }
       return null;
    }

    public boolean isPrecompileGeneration() {
      return isPrecompiledHeaderGeneration;
    }

}
