/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks.gcc;

import java.io.File;
import java.util.Vector;

import net.sf.antcontrib.cpptasks.CUtil;
import net.sf.antcontrib.cpptasks.compiler.LinkType;
import net.sf.antcontrib.cpptasks.compiler.Linker;

/**
 * Adapter for the GCC linker
 *
 * @author Adam Murdoch
 */
public final class GccLinker extends AbstractLdLinker
{
    private static final String[] objFiles = new String[]
      { ".o", ".a", ".lib", ".dll", ".so", ".sl" };
    private static final String[] libtoolObjFiles = new String[]
      { ".fo", ".a", ".lib", ".dll", ".so", ".sl" };
    private static final String[] discardFiles = new String[0];

    private static final GccLinker instance =
      new GccLinker("gcc", objFiles, discardFiles, "", "", false,null);
    private static final GccLinker dllLinker =
      new GccLinker("gcc", objFiles, discardFiles, "lib", ".so", false,
      new GccLinker("gcc", objFiles, discardFiles, "lib", ".so", true, null));


    private File[] libDirs;

    private GccLinker(String command, String[] extensions,
        String[] ignoredExtensions, String outputPrefix,
        String outputSuffix,
        boolean isLibtool, GccLinker libtoolLinker) {
        super(command, "-version", extensions, ignoredExtensions,
          outputPrefix, outputSuffix,
          isLibtool, libtoolLinker);
    }

    public static GccLinker getInstance() {
        return instance;
    }


    protected void addImpliedArgs(boolean debug,
      LinkType linkType, Vector args) {
      super.addImpliedArgs(debug, linkType, args);
      if (getIdentifier().indexOf("mingw") >= 0) {
        if (linkType.isSubsystemConsole()) {
          args.addElement("-mconsole");
        }
        if (linkType.isSubsystemGUI()) {
          args.addElement("-mwindows");
        }
      }
    }

    /**
     * Returns library path.
     *
     */
    public File[] getLibraryPath() {
      if(libDirs == null) {
        //
        //   construct gcc lib path from machine and version
        //
        StringBuffer buf = new StringBuffer("/lib/gcc-lib/");
        buf.append(GccProcessor.getMachine());
        buf.append('/');
        buf.append(GccProcessor.getVersion());
        //
        //   build default path from gcc and system /lib and /lib/w32api
        //
        String[] impliedLibPath = new String[] {
           buf.toString(), "/lib/w32api", "/lib" };

        //
        //     read gcc specs file for other library paths
        //
        String[] specs = GccProcessor.getSpecs();
        String[][] libpaths = GccProcessor.parseSpecs(specs,"*link:",
            new String[] { "%q" });
        String[] libpath;
        if(libpaths[0].length > 0) {
          libpath = new String[libpaths[0].length+3];
          int i = 0;
          for(; i < libpaths[0].length; i++) {
            libpath[i] = libpaths[0][i];
          }
          libpath[i++] = buf.toString();
          libpath[i++] = "/lib/w32api";
          libpath[i++] = "/lib";
        } else {
          //
          //   if a failure to find any matches then
          //      use some default values for lib path entries
          libpath = new String[] {
              "/usr/local/lib/mingw",
              "/usr/local/lib",
              "/usr/lib/w32api",
              "/usr/lib/mingw",
              "/usr/lib",
              buf.toString(),
              "/lib/w32api",
              "/lib" };
        }
        for (int i = 0; i < libpath.length; i++) {
          if (libpath[i].indexOf("mingw") >= 0) {
            libpath[i] = null;
          }
        }
        //
        //   if cygwin then
        //     we have to prepend location of gcc32
        //       and .. to start of absolute filenames to
        //       have something that will exist in the
        //       windows filesystem
        if(GccProcessor.isCygwin()) {
          GccProcessor.convertCygwinFilenames(libpath);
        }

        //
        //  check that remaining entries are actual directories
        //
        int count = CUtil.checkDirectoryArray(libpath);
        //
        //   populate return array with remaining entries
        //
        libDirs = new File[count];
        int index = 0;
        for (int i = 0; i < libpath.length; i++) {
          if (libpath[i] != null) {
            libDirs[index++] = new File(libpath[i]);
          }
        }
      }
      return libDirs;
    }


    public Linker getLinker(LinkType type) {
      if(type.isStaticLibrary()) {
        return GccLibrarian.getInstance();
      }
      if(type.isSharedLibrary()) {
        return dllLinker;
      }
      return instance;
    }


    private static String[] linkerOptions = new String[] {
        "-nostartfiles", "-nostdlib", "-s", "-static",
        "-shared", "-symbolic", "-Xlinker" };

    /**
     * Allows drived linker to decorate linker option.
     * Override by GccLinker to prepend a "-Wl," to
     * pass option to through gcc to linker.
     *
     * @param buf buffer that may be used and abused in the decoration process,
     * must not be null.
     * @param arg linker argument
     */
    public String decorateLinkerOption(StringBuffer buf, String arg) {
      String decoratedArg = arg;
      if(arg.length() > 1 && arg.charAt(0) == '-') {
        switch(arg.charAt(1)) {
            //
            //   passed automatically by GCC
            //
          case 'g':
          case 'f':
          case 'm':
          case 'O':
          case 'W':
          case 'l':
          case 'L':
          case 'u':
          break;

          default:
          boolean known = false;
          for (int i = 0; i < linkerOptions.length; i++) {
             if (linkerOptions[i].equals(arg)) {
                known = true;
                break;
             }
          }
          if (!known) {
             buf.setLength(0);
             buf.append("-Wl,");
             buf.append(arg);
             decoratedArg = buf.toString();
          }
          break;
        }
      }
      return decoratedArg;
    }
}
