/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2002 The Ant-Contrib project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Ant", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package net.sf.antcontrib.cpptasks.gcc;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;

import net.sf.antcontrib.cpptasks.CUtil;
import net.sf.antcontrib.cpptasks.compiler.CaptureStreamHandler;


/**
 * A add-in class for Gcc processors
 *
 *
 */
public class GccProcessor
{
    //   the results from gcc -dumpmachine
    private static String machine;
    //   the results from gcc -dumpversion
    private static String version;
    private static String[] specs;
    private GccProcessor() {
    }


    public static String getVersion() {
      if(version == null) {
        String[] args = new String[] { "gcc", "-dumpversion" };
        String[] cmdout = CaptureStreamHandler.run(args);
        if (cmdout.length == 0) {
          version = "noversion";
        } else {
          version = cmdout[0];
        }
      }
      return version;
    }


    public static String getMachine() {
      if(machine == null) {
        String[] args = new String[] { "gcc", "-dumpmachine" };
        String[] cmdout = CaptureStreamHandler.run(args);
        if (cmdout.length == 0) {
          machine = "nomachine";
        } else {
          machine = cmdout[0];
        }
      }
      return machine;
    }


    /**
     * Returns the contents of the gcc specs file.
     *
     * The implementation locates gcc.exe in the executable
     * path and then builds a relative path name from
     * the results of -dumpmachine and -dumpversion.
     * Attempts to use gcc -dumpspecs to provide this
     * information resulted in stalling on the Execute.run
     *
     * @returns contents of the specs file
     */
    public static String[] getSpecs() {
       if(specs == null) {
          File gccParent = CUtil.getExecutableLocation("gcc.exe");
          if(gccParent != null) {
            //
            //  build a relative path like
            //    ../lib/gcc-lib/i686-pc-cygwin/2.95.3-5/specs
            //
            StringBuffer buf = new StringBuffer("../lib/gcc-lib/");
            buf.append(getMachine());
            buf.append('/');
            buf.append(getVersion());
            buf.append("/specs");
            //
            //  resolve it relative to the location of gcc.exe
            //
            String relativePath = buf.toString();
            File specsFile = new File(gccParent,relativePath);
            //
            //  found the specs file
            //
            try {
              //
              //  read the lines in the file
              //
              BufferedReader reader = new BufferedReader(new FileReader(specsFile));
              Vector lines = new Vector(100);
              String line = reader.readLine();
              while (line != null) {
                lines.addElement(line);
                line = reader.readLine();
              }
              specs = new String[lines.size()];
              lines.copyInto(specs);
            } catch(IOException ex) {
            }
          }
        }
        if(specs == null) {
          specs = new String[0];
        }
        return specs;
    }

    /**
     * Converts absolute Cygwin file or directory names
     * to the corresponding Win32 name.
     *
     * @param names array of names, some elements may be null,
     * will be changed in place.
     */
    public static void convertCygwinFilenames(String[] names) {
      if(names == null) {
        throw new NullPointerException("names");
      }
      File gccDir = CUtil.getExecutableLocation("gcc.exe");
      if (gccDir != null) {
        String prefix = gccDir.getAbsolutePath() + "/..";
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < names.length; i++) {
          String name = names[i];
          if (name != null && name.length() > 1 && name.charAt(0) == '/') {
            buf.setLength(0);
            buf.append(prefix);
            buf.append(name);
            names[i] = buf.toString();
          }
        }
      }
    }



    /**
     *
     * Parses the results of the specs file for a
     * specific processor and options
     *
     * @param specsContent Contents of specs file as returned from getSpecs
     * @param specSectionStart start of spec section, for example "*cpp:"
     * @param options command line switches such as "-istart"
     */
    public static String[][] parseSpecs(String[] specsContent,
      String specSectionStart, String[] options) {
      if (specsContent == null) {
        throw new NullPointerException("specsContent");
      }
      if (specSectionStart == null) {
        throw new NullPointerException("specSectionStart");
      }
      if (options == null) {
        throw new NullPointerException("option");
      }
      String[][] optionValues = new String[options.length][];

      StringBuffer optionValue = new StringBuffer(40);

      for (int i = 0; i < specsContent.length; i++) {
        String specLine = specsContent[i];
       //
        //   if start of section then start paying attention
        //
        if(specLine.startsWith(specSectionStart)) {
          Vector[] optionVectors = new Vector[options.length];
          for (int j = 0; j < options.length; j++) {
            optionVectors[j] = new Vector(10);
          }
          //
          //  go to next line and examine contents
          //     and repeat until end of file
          //
          for (i++; i < specsContent.length; i++) {
            specLine = specsContent[i];
            for (int j = 0; j < options.length; j++) {
              int optionStart = specLine.indexOf(options[j]);
              while (optionStart >= 0) {
                optionValue.setLength(0);
                //
                //   walk rest of line looking for first non whitespace
                //       and then next space
                boolean hasNonBlank = false;
                int k = optionStart + options[j].length();
                for (; k < specLine.length(); k++) {
                  //
                  //  either a blank or a "}" (close of conditional)
                  //    section will end the path
                  //
                  if (specLine.charAt(k) == ' ' || specLine.charAt(k) == '}') {
                    if(hasNonBlank) {
                      break;
                    }
                  }
                  else {
                    hasNonBlank = true;
                    optionValue.append(specLine.charAt(k));
                  }
                }
                //
                //  transition back to whitespace
                //     value is over, add it to vector
                if (hasNonBlank) {
                  optionVectors[j].addElement(optionValue.toString());
                }
                //
                //  find next occurance on line
                //
                optionStart = specLine.indexOf(options[j],k);
              }
            }
          }
          //
          //   copy vectors over to option arrays
          //
          for (int j = 0; j < options.length; j++) {
            optionValues[j] = new String[optionVectors[j].size()];
            optionVectors[j].copyInto(optionValues[j]);
          }
        }
      }
      //
      //   fill in any missing option values with
      //      a zero-length string array
      for (int i = 0; i < optionValues.length; i++) {
        String[] zeroLenArray = new String[0];
        if (optionValues[i] == null) {
          optionValues[i] = zeroLenArray;
        }
      }
      return optionValues;
    }

    /**
     * Determines if task is running with cygwin
     *
     * @return true if cygwin was detected
     */
    public static boolean isCygwin() {
      return getMachine().indexOf("cygwin") > 0;
    }

    public static String[] getOutputFileSwitch(String letter,String outputFile) {
      StringBuffer buf = new StringBuffer();
      if (outputFile.indexOf(' ') >= 0) {
        buf.append('"');
        buf.append(outputFile.replace('\\','/'));
        buf.append('"');
      } else {
        buf.append(outputFile.replace('\\','/'));
      }
      String[] retval = new String[] { letter, buf.toString() };
      return retval;
    }

    private static int addLibraryPatterns(
      String[] libnames,
      StringBuffer buf,
      String prefix,
      String extension,
      String[] patterns,
      int offset) {
      for(int i = 0; i < libnames.length; i++) {
        buf.setLength(0);
        buf.append(prefix);
        buf.append(libnames[i]);
        buf.append(extension);
        patterns[offset + i] = buf.toString();
      }
      return offset + libnames.length;
    }

    private static boolean isHPUX() {
      String osname = System.getProperty("os.name").toLowerCase();
      if(osname.indexOf("hp") >= 0 && osname.indexOf("ux") >= 0) {
        return true;
      }
      return false;
    }

    public static String[] getLibraryPatterns(String[] libnames) {
      StringBuffer buf = new StringBuffer();
      String[] patterns = new String[libnames.length * 2];
      int offset = addLibraryPatterns(libnames,buf,
        "lib",".a",patterns,0);
      if(isHPUX()) {
        offset = addLibraryPatterns(libnames,buf,
          "lib",".sl",patterns,offset);
      } else {
        offset = addLibraryPatterns(libnames,buf,
          "lib",".so",patterns,offset);
      }
      return patterns;
    }

    public static boolean isCaseSensitive() {
      return true;
    }

}
